;** Fitxer: salvrest.asm
;** Data: Agost 2002
;** Autors: Sergio Blanco Cuaresma
;**         Jonathan Hernndez Velasco
;**         Valentin Rey Gmez
;** Descripci: Mdul en assamblador amb les funcions necessaries del kernel


public _iniPila
public _salvarContext
public _restaurarContext
public _obtenirAX
public _canviarAX
public _llegirTecla


_DATA   segment word public 'DATA'
_DATA   ends

_BSS    segment word public 'BSS'
_BSS    ends

DGROUP  group   _DATA, _BSS

assume cs:SALVREST_CODI, ds:DGROUP


SALVREST_CODI segment byte public 'CODE'

; Inicialitza una pila a la zona de memoria que li diguem
; guardant primer el flags i l'adrea de retorn del proces al
; que pertany la pila
_iniPila proc far
    push bp
    mov bp, sp

    mov cx, ss      ; guardem la posici de la pila actual
    mov bx, sp

    mov ax, ss:[bp+6]      ; segment del procs --> AX
    mov dx, ss:[bp+8]      ; desplaament del procs --> DX

    mov sp, ss:[bp+12]
    mov ss, ss:[bp+10]      ; redireccionem cap a la nostra pila

    pushf               ; fiquem els flags a la pila
    pop bp		; ens assegurem que el flag d'interrupcio ha d'estar activat
    or bp, 200h
    push bp
    
    push dx             ; fiquem segment del procs
    push ax             ; fiquem desplaament del procs

    push 0      ; bp
    push 1      ; cx
    push 2      ; ax
    push 3      ; bx
    push 4      ; si
    push 5      ; di
    push ds
    push es
    push 6      ; dx

    mov dx, ss      ; retornem la posici final de la pila
    mov ax, sp

    mov ss, cx      ; restaurem la pila
    mov sp, bx

    pop bp
    ret
_iniPila endp

; Guarda el context a la pila de l'ultim proces que s'estaba executant
; com que ha sigut interromput per una interrupcio, l'adrea a on s'ha quedat
; l'execuci esta a la pila, per tant, aprofitarem aquesta situaci
; * A la pila tindrem l'adrea de retorn de la RSI i la del proces interromput
_salvarContext proc far
    push bx
    mov bx, sp
    xchg ax, ss:[bx+2]	; guardem ax a la pila i agafem adrea de retorn de la RSI
    xchg cx, ss:[bx+4]  ; idem anterior

    push si
    push di
    push ds
    push es
    push dx

    push cx     ; restaurem l'adrea de retorn
    push ax

    mov ax, DGROUP
    mov ds, ax
    mov es, ax

    mov dx, ss      ; retornem el ponter cap a la nostra pila
    mov ax, sp
    add ax, 4       ; saltem l'adrea de retorn
    ret
_salvarContext endp

; Restaurem context i amb iret saltarem al punt on va ser interromput el proces
_restaurarContext proc far
    add sp, 4       ; saltem l'adrea de retorn
    pop ax          ; ens situem en l'adrea de la pila
    pop ss          ; que se li ha passat per parmetre
    mov sp, ax

    pop dx          ; restaurem els registres
    pop es
    pop ds
    pop di
    pop si
    pop bx
    pop ax
    pop cx
    pop bp
  
    iret
_restaurarContext endp


; Retorna el valor actual del registre AX
_obtenirAX proc far
    ret
_obtenirAX endp


; Util per actualitzarBloquejats del teclat, canvia el registre AX d'un pcb determinat
_canviarAX proc far
    push bp
    mov bp, sp
    push es
    push bx
    push cx

    mov bx,[bp+6]
    mov es,[bp+8]
    mov cx,[bp+10]
    mov word ptr es:[bx+12],cx

    pop cx
    pop bx
    pop es
    pop bp
    ret    
_canviarAX endp


; Retorna la tecla que estigui almacenada (s'ha de cridar des de la RSI del teclat, no des de cap altre banda)
_llegirTecla PROC FAR
	in      al, 60h         ; AL <- "Scancode" Tecla polsada
        test    al, 10000000b           ; Make/Break?
        jz     NoEsBreak    
        mov	al, 1		;Ignorem Break, fiquem com a senyal un 1
NoEsBreak:
        mov	ah, 0
	ret
_llegirTecla ENDP


SALVREST_CODI ends

        end
