/*
** Fitxer: main.c
** Data: Agost 2002
** Autors: Sergio Blanco Cuaresma
**         Jonathan Hernndez Velasco
**         Valentin Rey Gmez
** Descripci: Mdul amb el codi de les rutines d'interrupci multiplexar, trap i teclat,
**             a ms de totes les rutines internes del nucli i el programa principal.
**
*/

#include "ec.h"
#include "estruct.h"
#include "pantalla.h"
#include "cues.h"
#include "nucli.h"
#include "procs.h"
#include "crides.h"
#include "salvrest.h"


WORD flags;
struct cua cuaRUN, cuaRDY1, cuaRDY2, cuaBLK, cuaFREE, cuaSLEEP; // Cua run, ready 1 i 2, blocked, pcbs lliures respectivament i adormits
struct cua *pRUN, *pRDY1, *pRDY2, *pBLK, *pFREE, *pSLEEP;  // Apunten als seus homlegs anteriors
// Per restaurar les velles RSIs
void (*rsi_ck)();
void (*rsi_tr)();
void (*rsi_kyb)();
// Reserva de PCBs
tPCB pcbs[NPROC];
// Identificador que s'usar per la creaci de processos, s'anira incrementant
char identificadorProc='A';
tPCB *procRun;	// Apunta al proces en execucio, til per acces fcil
int Quantum=QUANTUM_CUA1;
int cuaProcedent=1;	// Indica de quina cua de Ready procedeix el proc en execucio (1=cua prioritaria, 2=cua menys prioritaria)
tSemafor Semafors[NSEM];
int SemaforsLliures=NSEM;
tSemafor SemaforPantalla;	// Semafor per accedir a la pantalla (separat del altres semafor per evitar que un usuari accedeixi)
char bufferTeclat[BUFFERTECLAT];
int tecles=0;	// Nombre de tecles pitjades
// Smbol que s'usar per mostrar informaci per pantalla
char simbol[]="|/-\\";

void iniCues(void);
void multiplexar (void);
void canviarVectorInt(void);
void restaurarVectorInt(void);
tPCB *scheduler(void);
void dispatcher(tPCB *proc);
void actualitzarDormits(void);
void Shell(void);
void killall(void);
void llistarCues(void);
void ajuda(void);
void clearscreen(void);


// Inclou tots els pcbs a la cua de pcbs lliures
// i inicialitza la resta de cues
void iniCues(void){
  tPCB *punter_pcb;
  int i;
 
  pFREE=&cuaFREE;
  pBLK=&cuaBLK;
  pRDY1=&cuaRDY1;
  pRDY2=&cuaRDY2;
  pRUN=&cuaRUN;
  pSLEEP=&cuaSLEEP;
  // Inicializamos la cua de pcbs libres
  Inicialitzar_cua(pFREE);
  for (i=0;i<NPROC;i++) {
 	pcbs[i].IdPCB = i;
 	pcbs[i].Estat = FREE;
 	punter_pcb = &(pcbs[i]);
 	Encuar (pFREE, (struct item *)punter_pcb);
  };
  Inicialitzar_cua(pRDY1);
  Inicialitzar_cua(pRDY2);
  Inicialitzar_cua(pBLK);
  Inicialitzar_cua(pRUN);
  Inicialitzar_cua(pSLEEP);
}

// Inicialitzem el semafors i les seves cues
void iniSemafors(void)
{
 int i;
 for(i=0;i<NSEM;i++)
 {
  Semafors[i].id=NSEM;	// Aix simbolitza que no esta inicialitzat
  Semafors[i].pBLK=&Semafors[i].cuaBLK;
 }
 SemaforPantalla.pBLK=&SemaforPantalla.cuaBLK;
 SemaforPantalla.id=0;
 SemaforPantalla.valor=1;
}

// Carrega proces Shell i queda inactiu (nomes mostra per pantalla un simbol)
void Null(void)
{
  static int i;
  static char c;
  

  clearscreen();
  i=0;
  EscriurePantalla("Null(",5,24,57,NORMAL);
  c=QuiSoc();
  EscriurePantalla(&c,1,24,62,NORMAL);
  EscriurePantalla(")",1,24,63,NORMAL);
 
  CrearProces(Shell,9);
  for (;;){
    EscriurePantalla(&simbol[(i=(i+1)%4)], 1, 24,65,NORMAL);
  }
}


void Shell(void)
{
 static char opcio;
 static int llegits;
 static int i;

 // Decoracio
 EscriurePantalla("Shell(",6,24,45,NORMAL);
 opcio=QuiSoc();
 EscriurePantalla(&opcio,1,24,51,NORMAL);
 EscriurePantalla(")",1,24,52,NORMAL);
 for (i=0;i<80;i++) EscriurePantalla("_",1,23,i,MARRO);
 EscriurePantalla("->",2,24,0,GROC);
 
 // Nomes entrar, anar directes a l'ajuda
 ajuda();
 for (;;)
 {
  EscriurePantalla(&simbol[(i=(i+1)%4)], 1, 24,54,NORMAL);
  llegits=LlegirTeclat(&opcio,1);
  EscriurePantalla("   ",3,24,2,NORMAL);
  EscriurePantalla(&opcio,llegits,24,3,NORMAL);
  switch (opcio)
  {
   case 'R': EscriurePantalla("Joc de proves #",15,22,1,NORMAL);
     llegits=LlegirTeclat(&opcio,1);
     EscriurePantalla(&opcio,llegits,24,4,NORMAL);
     switch (opcio) {
       case '1': killall(); clearscreen();
                 CrearProces(jp1,8);
		 break;
       case '2': killall(); clearscreen();
                 CrearProces(jp2,5);
		 break;
       case '3': killall(); clearscreen();
                 CrearProces(jp3,5);
		 break;
       case '4': killall(); clearscreen();
                 CrearProces(jp4,5);
		 break;
       case '5': killall(); clearscreen();
                 CrearProces(jp5,5);
		 break;
       case '6': killall(); clearscreen();
                 CrearProces(jp6,5);
		 break;
       case '7': killall(); clearscreen();
                 CrearProces(jp7,5);
		 break;
       case '8': killall(); clearscreen();
                 CrearProces(jp8,5);
		 break;
       case '9': killall(); clearscreen();
                 CrearProces(jp9,5);
		 break; 
     }
     EscriurePantalla("   ",3,24,2,NORMAL);
     break;
   case 'K': clearscreen();
	     killall();
	     break;
   case 'N': clearscreen();
	     break;
   case 'L': llistarCues();
   	     break;
   case 'A': killall();		// Ho matem tot pq no molesti a l'hora de llegir
   	     ajuda();
	     break;
   case 'Q': flags=Inhibir();
             CanviarVectInt(0x08,rsi_ck);
             CanviarVectInt(0x40,rsi_tr);
             CanviarVectInt(0x09,rsi_kyb);
             Desinhibir(flags);
             exit(0);
	     break;
   default: break;
  }
 }
}

// Borra la part superior de la pantalla
void clearscreen(void){
  int i,j;
  
  for (i=0;i<80;i++) {
    for (j=0;j<23;j++) {
      EscriurePantalla(" ",1,j,i,NORMAL);
    }
  }
}


// Visualitza l'ajuda, no finalitza fins que es pren una tecla
void ajuda(void){
  int i,j;
  char c;
  
  for (i=0;i<80;i++) {
    for (j=0;j<23;j++) {
      EscriurePantalla(" ",1,j,i,NORMAL);
    }
  }
  EscriurePantalla("Ajuda",5,1,1,VERMELL);
  EscriurePantalla("-----",5,2,1,VERMELL);
  EscriurePantalla("R1",2,4,1,GROC);EscriurePantalla("Comprovacio del Round Robin",27,4,4,NORMAL);
  EscriurePantalla("R2",2,5,1,GROC);EscriurePantalla("Comprovacio del Quantum",23,5,4,NORMAL);
  EscriurePantalla("R3",2,6,1,GROC);EscriurePantalla("Comprovacio de les prioritats",29,6,4,NORMAL);
  EscriurePantalla("R4",2,7,1,GROC);EscriurePantalla("Comprovacio de DormirProces",27,7,4,NORMAL);
  EscriurePantalla("R5",2,8,1,GROC);EscriurePantalla("Comprovacio de les prioritats i apropiacio",42,8,4,NORMAL);
  EscriurePantalla("R6",2,9,1,GROC);EscriurePantalla("Comprovacio dels semafors",25,9,4,NORMAL);
  EscriurePantalla("R7",2,10,1,GROC);EscriurePantalla("Comprovacio del teclat",23,10,4,NORMAL);
  EscriurePantalla("R8",2,11,1,GROC);EscriurePantalla("Comprovacio de la reutilitzacio dels PCB",41,11,4,NORMAL);
  EscriurePantalla("R9",2,12,1,GROC);EscriurePantalla("Problema dels filosofs",23,12,4,NORMAL);
  EscriurePantalla("K",1,13,1,GROC);EscriurePantalla("KillAll - Mata tots els processos",33,13,4,NORMAL);
  EscriurePantalla("N",1,14,1,GROC);EscriurePantalla("Neteja la pantalla",18,14,4,NORMAL);
  EscriurePantalla("L",1,15,1,GROC);EscriurePantalla("Llista els identificadors de les cues",37,15,4,NORMAL);
  EscriurePantalla("Q",1,16,1,GROC);EscriurePantalla("Quit (sortir)",13,16,4,NORMAL);
  EscriurePantalla("A",1,18,1,MARRO);EscriurePantalla("Ajuda (fa un killall per evitar molesties)",42,18,4,NORMAL);
  EscriurePantalla("Pressioni qualsevol tecla per continuar",39,22,1,NORMAL);
  LlegirTeclat(&c,1);
  for (i=0;i<80;i++) {
    for (j=0;j<23;j++) {
      EscriurePantalla(" ",1,j,i,NORMAL);
    }
  }
}

// Carrega el proces null (el qual carregara el proces Shell i despres es quedara inactiu)
void CrearProcNull(void){
 tPCB *pcbAuxiliar;
 WORD ss, sp;
 
 // Creem proces NULL
 pcbAuxiliar=Primer(pFREE);
 pcbAuxiliar->IdProces = identificadorProc;
 identificadorProc=(identificadorProc+1)%256;
 pcbAuxiliar->Estat = RDY;
 pcbAuxiliar->Prioritat = 0;
 pcbAuxiliar->QSencer[0] = TRUE;
 pcbAuxiliar->QSencer[1] = TRUE;
 ss = Segment(pcbAuxiliar->pila);
 sp = Desplacament(&(pcbAuxiliar->pila[MIDA_PILA-1]));
 pcbAuxiliar->SS_SP=iniPila(Null,ss,sp);
 procRun = pcbAuxiliar;
 Inserir (pRDY1, (struct item *) pcbAuxiliar, pcbAuxiliar->Prioritat);

 Quantum=QUANTUM_CUA1;
}


// Llista el contingut (el id proces) de totes les llistes, les que no contenen res sn ignorades
void llistarCues(void){
  tPCB *pcbAuxiliar;
  int i,bucle;
  static c=22;
  char idsem;
  
 WORD flags;

 flags=Inhibir();
  
  // Borrem nomes la zona en la que hem escrit en una crida anterior
  while (c<=22){
    EscriureString(c,0,"                                                                               ",NORMAL);
    c++;
  }
  
  c=22;  //Columna
  
  // Visualitzem cua de Run
   EscriureString(c,1,"Cua RUN:",NORMAL);
   EscriureCar(c,10,procRun->IdProces,GROC);
   c--;

  // Visualitzem cua de Ready1
  if (!Buida(pRDY1)) {
    EscriureString(c,1,"Cua RDY1:",NORMAL);
    pcbAuxiliar=(tPCB *)pRDY1->primer;
  
    i=11;
    while ((struct TipusPCB *)pRDY1 != pcbAuxiliar) {
  	EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	i++;
  	i++;
    }
    c--;
  }
  
  // Visualitzem cua de Ready 2
  if (!Buida(pRDY2)) {
    EscriureString(c,1,"Cua RDY2:",NORMAL);
    pcbAuxiliar=(tPCB *)pRDY2->primer;
  
    i=11;
    while ((struct TipusPCB *)pRDY2 != pcbAuxiliar) {
  	EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	i++;
  	i++;
    }
    c--; 
  }
  
  // Visualitzem cua de Bloquejats
  if (!Buida(pBLK)) {
    EscriureString(c,1,"Cua BLK:",NORMAL);
    pcbAuxiliar=(tPCB *)pBLK->primer;
  
    i=10;
    while ((struct TipusPCB *)pBLK != pcbAuxiliar) {
  	EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	i++;
  	i++;
    }
    c--;
  }
  
  // Visualitzem cua de Dormits
  if (!Buida(pSLEEP)) {
    EscriureString(c,1,"Cua SLEEP:",NORMAL);
    pcbAuxiliar=(tPCB *)pSLEEP->primer;
  
    i=12;
    while ((struct TipusPCB *)pSLEEP != pcbAuxiliar) {
  	EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	i++;
  	i++;
    }
    c--;
  }
  
  // Visualitzem les cues dels semafors actius
  for (bucle=0;bucle<NSEM;bucle++){
    if (Semafors[bucle].id != NSEM){
      if (!Buida(Semafors[bucle].pBLK)) {
        EscriureString(c,1,"Cua SLEEP Semafor  :",NORMAL);
        idsem=Semafors[bucle].id;
        EscriureCar(c,19,48+Semafors[bucle].id,NORMAL);
        pcbAuxiliar=(tPCB *)Semafors[bucle].pBLK->primer;
        i=22;
        while ((struct TipusPCB *)Semafors[bucle].pBLK != pcbAuxiliar) {
  	  EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	  pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	  i++;
  	  i++;
        }
        c--;
      }
    }
  }
  
  // Visualitzem cua del Semafor de pantalla
  if (!Buida(SemaforPantalla.pBLK)){
    EscriureString(c,1,"Cua Semafor Pantalla:",NORMAL);
    pcbAuxiliar=(tPCB *)SemaforPantalla.pBLK->primer;
  
    i=22;
    while ((struct TipusPCB *)SemaforPantalla.pBLK != pcbAuxiliar) {
  	EscriureCar(c,i,pcbAuxiliar->IdProces,GROC);
  	pcbAuxiliar=pcbAuxiliar->Fil.seguent;
  	i++;
  	i++;
    }
    c--;
   }
  
  Desinhibir(flags);
}

// Mata tos els procesos, excepte el Null i el Shell (aquest ultim esta a la cua de RUN que no la toquem)
void killall(void){
 tPCB *pcbAuxiliar, *pcbNull, *pcbShell;
 int nullTrue=0;
 int i;
 WORD flags;

 flags=Inhibir();
 
 // Eliminem tots els semafors ocupats i buidem les seves cues
 SemaforsLliures=NSEM;
 for (i=0;i<NSEM;i++){
   if (Semafors[i].id!=NSEM){
     while (!Buida(Semafors[i].pBLK)){
       pcbAuxiliar=(tPCB *)Primer(Semafors[i].pBLK);
       Encuar(pFREE,(struct item *)pcbAuxiliar);
     }
     Semafors[i].id=NSEM;
   }
 }
 
 nullTrue=0;
 // Buidem la cua de Ready1
 while (!Buida(pRDY1))
 {
  pcbAuxiliar=(tPCB *)Primer(pRDY1);
  // alliberem el PCB si el proces no es el Null
  if (pcbAuxiliar->IdProces=='A'){		// Mantenim proces null (i shell tb pq es el que esta en cua RUN)
    pcbNull=pcbAuxiliar;
    nullTrue=1;
  } else {
    Encuar(pFREE,(struct item *)pcbAuxiliar);
  }
 }
 if (nullTrue){
   Inserir(pRDY1, (struct item *)pcbNull, pcbNull->Prioritat);
 }

 nullTrue=0;
 // Buidem la cua de Ready2
 while (!Buida(pRDY2))
 {
  pcbAuxiliar=(tPCB *)Primer(pRDY2);
  // alliberem el PCB si el proces no es el Null
  if (pcbAuxiliar->IdProces=='A'){		// Mantenim proces null (i shell tb pq es el que esta en cua RUN)
    pcbNull=pcbAuxiliar;
    nullTrue=1;
  } else {
    Encuar(pFREE,(struct item *)pcbAuxiliar);
  }
 }
 if (nullTrue){
   Inserir(pRDY2, (struct item *)pcbNull, pcbNull->Prioritat);
 }
 
 // Buidem la cua de Blocked
 while (!Buida(pBLK))
 {
  pcbAuxiliar=(tPCB *)Primer(pBLK);
  if ((pcbAuxiliar->IdProces=='B')){	// Mantenim proces Shell que podria estar bloquejat per teclat
    Encuar(pRDY1, (struct item *)pcbAuxiliar);
  }else{
    Encuar(pFREE,(struct item *)pcbAuxiliar);
  }
 }

 // Buidem la cua de Dormits
 while (!Buida(pSLEEP))
 {
  pcbAuxiliar=(tPCB *)Primer(pSLEEP);
  Encuar(pFREE,(struct item *)pcbAuxiliar);
 }

 // Buidem la cua del semafor de pantalla
 while (!Buida(SemaforPantalla.pBLK))
 {
  pcbAuxiliar=(tPCB *)Primer(SemaforPantalla.pBLK);
  Encuar(pFREE,(struct item *)pcbAuxiliar);
 }
 
 identificadorProc='C';	// L'identificador A i B agafats per null i shell respectivament
 Desinhibir(flags);
}


// Fa un wait al semafor de la pantalla que el tractem de forma independent a la resta
// per evitar que l'usuari tingui acces
int WaitSPantalla()
{
  static tPCB *pcbAuxiliar;
  static WORD flags;

  flags=Inhibir();

  // si el semafor esta utilitzant-se
  if (SemaforPantalla.valor==0)
  {
   Desinhibir(flags);
   // El treiem de Run per ficar-ho a cua BLK
   //procRun->SS_SP=salvarContext();  // No cal, es cridara desde una crida sist. i ja s'haura guardat al trap
   pcbAuxiliar=(struct TipusPCB *)Primer(pRUN);     // treiem el proces de la cua de Run
   pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
   pcbAuxiliar->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
   pcbAuxiliar->Estat=BLK;
   Encuar(SemaforPantalla.pBLK,(struct item *)pcbAuxiliar);
   // Agafem nou proces
   if (!Buida(pRDY1)){
     cuaProcedent=1;
     Quantum=QUANTUM_CUA1;
     dispatcher((tPCB *)Primer(pRDY1));
   } else {
     Quantum=QUANTUM_CUA2;
     cuaProcedent=2;
     dispatcher((tPCB *)Primer(pRDY2));
   }
  } else {
   SemaforPantalla.valor--;
   Desinhibir(flags);
  }
  return(0);
}

// Fa un signal al semafor de la pantalla que el tractem de forma independent a la resta
// per evitar que l'usuari tingui acces
int SignalSPantalla()
{
  static WORD flags;
  static tPCB *pcbAuxiliar;
  static tPCB *pcbExe;

  flags=Inhibir();
  if (Buida(SemaforPantalla.pBLK)) {
    SemaforPantalla.valor++;
  } else {
    Desinhibir(flags);
    pcbAuxiliar=(tPCB *)Primer(SemaforPantalla.pBLK);
    if (procRun->Prioritat >= pcbAuxiliar->Prioritat) {		// No hem de fer apropiacio
        Inserir(pRDY1,(struct item *)pcbAuxiliar,pcbAuxiliar->Prioritat);
    } else {		// Hem de fer apropiacio
    	Desinhibir(flags);
        //procRun->SS_SP=salvarContext();  // No cal, es cridara desde una crida sist. i ja s'haura guardat al trap
        pcbExe = (tPCB *)Primer(pRUN);
        pcbExe->Estat = RDY;
        pcbExe->QSencer[0] = pcbExe->QSencer[1];	// Historic d'execucions amb quantum complet
        pcbExe->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
        // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
        // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
        Inserir (pRDY1, (struct item*)pcbExe, pcbExe->Prioritat);
        cuaProcedent=1;
        Quantum = QUANTUM_CUA1;
        dispatcher(pcbAuxiliar);
    }
  }
  Desinhibir(flags);
  return(0);
}


// S'encarrega dels canvis de context
void multiplexar (void)
{
  static int j=0,i=0;

  procRun->SS_SP=salvarContext();
  // Visualitzem informaci per pantalla
  EscriureString(24,13,"Running",VERMELL);
  EscriureCar(24,21,procRun->IdProces,GROC);
  EscriureString(24,23,"CuaRDY",VERMELL);
  EscriureCar(24,30,48+cuaProcedent,GROC);
  EscriureString(24,32,"Prioritat",VERMELL);
  EscriureCar(24,42,48+procRun->Prioritat,GROC);
  EscriureString(24,68,"Quantum",VERMELL);
  EscriureCar(24,76,simbol[(i=(i+1)%4)],GROC);
  if (Quantum<10) {		// Nomes funciona be per quantums entre 0 i 99
    EscriureCar(24,79,48+Quantum,INVERSE);
    EscriureCar(24,78,' ',INVERSE);
  } else if (Quantum<100){
    EscriureCar(24,78,48+ (Quantum-(Quantum%10))/10,INVERSE);
    EscriureCar(24,79,48+ (Quantum%10),INVERSE);
  } else {
    EscriureCar(24,78,'*',INVERSE);
    EscriureCar(24,79,'*',INVERSE);
  }
  
    // actualitzem la cua de Dormirs
  if (!Buida(pSLEEP)) actualitzarDormits();
  
  if (Quantum == 0) {
    // Despres d'aquesta func s'efectuara el canvi de context real
    dispatcher(scheduler());
  }
  Quantum--;
  EOI();
  restaurarContext(procRun->SS_SP);
}

// Controla cua de Bloquejats, si hi ha algun que ja pot sortir li dona pas
void actualitzarBloquejats(void)
{
 tPCB *pcbExe, *procBLK;
 tPCB *pcbAuxiliar;
 int i;

  procBLK=(tPCB *)pBLK->primer;
  if (procBLK->llegir==tecles || bufferTeclat[tecles-1]==13) {	// Si ha sigut un ENTER o ja hem llegit totes les tecles
   //Traspasem buffer al pcb
   for (i=0;i<procBLK->llegir;i++) {
     procBLK->buffer[i]=bufferTeclat[i];
   }
   
   // Fem que la crida a sistema retorni el nombre de tecles pitjades modificant
   // el registre AX salvat amb el context
   canviarAX(procBLK->SS_SP,tecles);
   tecles=0;
   pcbAuxiliar=(tPCB *)Primer(pBLK);
   // Mirem si s'ha de fer apropiaci
   if (procRun->Prioritat >= pcbAuxiliar->Prioritat) {
   	pcbAuxiliar->Estat=RDY;
        Inserir(pRDY1,(struct item *)pcbAuxiliar,pcbAuxiliar->Prioritat);
   } else {
        //procRun->SS_SP=salvarContext();	// No cal, ve de la rsi del teclat el qual ja l'ha guardat
        pcbExe = (tPCB *)Primer(pRUN);
        pcbExe->Estat = RDY;
        pcbExe->QSencer[0] = pcbExe->QSencer[1];	// Historic d'execucions amb quantum complet
        pcbExe->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
        // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
        // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
        Inserir (pRDY1, (struct item*)pcbExe, pcbExe->Prioritat);
        cuaProcedent=1;
        Quantum = QUANTUM_CUA1;
        dispatcher(pcbAuxiliar);
   }
  }
 
}


// RSI del teclat
void teclat(void)
{
 static char c;

 procRun->SS_SP=salvarContext();
 c=llegirTecla();	// Llegeix el ScanCode de la tecla pulsada, retorna 1 si es BREAK
 if (c!=1)              // Si es MAKE, traducci a ASCII i tractament
 {
   c=ASCII(c);
   bufferTeclat[tecles]=c;
   tecles++;
   if (!Buida(pBLK)) {
     actualitzarBloquejats();
   }
 }

 // Restaurem el proces que ara toca ser RUN (potser es el mateix)
 EOI();
 restaurarContext(procRun->SS_SP);
}


// Controla cua de SLEEP, desperta els que siguin necessaris
void actualitzarDormits(void){
  static tPCB *pcbAuxiliar;
  static tPCB *pcbExe;
  int idUltim;

  pcbAuxiliar=(struct TipusPCB *)pSLEEP->ultim;
  idUltim=pcbAuxiliar->IdProces;
  pcbAuxiliar=(struct TipusPCB *)Primer(pSLEEP);

  while (pcbAuxiliar->IdProces!=idUltim)
  {
    (pcbAuxiliar->Dormir)--;
    if (pcbAuxiliar->Dormir==0)  // si ha acabat el temps de Dormir el fiquem a Ready
    {
      if (procRun->Prioritat >= pcbAuxiliar->Prioritat) {
      	pcbAuxiliar->Estat=RDY;
        Inserir(pRDY1,(struct item *)pcbAuxiliar,pcbAuxiliar->Prioritat);
      } else {
        //procRun->SS_SP=salvarContext();	// No cal, ve del multiplexar el qual ja l'ha guardat
        pcbExe = (tPCB *)Primer(pRUN);
        pcbExe->Estat = RDY;
        pcbExe->QSencer[0] = pcbExe->QSencer[1];	// Historic d'execucions amb quantum complet
        pcbExe->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
        // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
        // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
        Inserir (pRDY1, (struct item*)pcbExe, pcbExe->Prioritat);
        cuaProcedent=1;
        Quantum = QUANTUM_CUA1;
        dispatcher(pcbAuxiliar);
      }
    } else {
      Encuar(pSLEEP,(struct item *)pcbAuxiliar);
    }
    pcbAuxiliar=(struct TipusPCB *)Primer(pSLEEP);
  }

 // hem de tractar l'ultim element de la cua apart
  (pcbAuxiliar->Dormir)--;
  if (pcbAuxiliar->Dormir==0)  // si ha acabat el temps de Dormir el fiquem a Ready
  {
    if (procRun->Prioritat >= pcbAuxiliar->Prioritat) {
      pcbAuxiliar->Estat=RDY;
      Inserir(pRDY1,(struct item *)pcbAuxiliar,pcbAuxiliar->Prioritat);
    } else {
      //procRun->SS_SP=salvarContext();
      pcbExe = (tPCB *)Primer(pRUN);
      pcbExe->Estat = RDY;
      pcbExe->QSencer[0] = pcbExe->QSencer[1];	// Historic d'execucions amb quantum complet
      pcbExe->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
      // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
      // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
      Inserir (pRDY1, (struct item*)pcbExe, pcbExe->Prioritat);
      cuaProcedent=1;
      Quantum = QUANTUM_CUA1;
      dispatcher(pcbAuxiliar);
    }
  } else {
    Encuar(pSLEEP,(struct item *)pcbAuxiliar);
  }
}

// Agafa proces de la cua de Ready que ha de passar a Run
tPCB *scheduler(void){
  static tPCB *pcbAuxiliar;
  
  pcbAuxiliar = (tPCB *)Primer(pRUN);
  
  // S'haura de comprovar que si es troba en estat sleeping, blocked o fi no s'ha d'insertar a ready
  pcbAuxiliar->Estat = RDY;
  pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
  pcbAuxiliar->QSencer[1] = TRUE;			// True, ha utilitzat tot el quantum
  // Eleccio de la cua
  if ((pcbAuxiliar->QSencer[0]) && (pcbAuxiliar->QSencer[1])){
    Inserir (pRDY2, (struct item*)pcbAuxiliar, pcbAuxiliar->Prioritat);
  } else if ((pcbAuxiliar->QSencer[1]==FALSE) && (cuaProcedent==2)) {
    Inserir (pRDY1, (struct item*)pcbAuxiliar, pcbAuxiliar->Prioritat);
  } else {	// Nomes falten proc que aniran a les cues d'on provenien
    if (cuaProcedent==1){
      Inserir (pRDY1, (struct item*)pcbAuxiliar, pcbAuxiliar->Prioritat);
    } else {
      Inserir (pRDY2, (struct item*)pcbAuxiliar, pcbAuxiliar->Prioritat);
    }
  }
  
  if (!Buida(pRDY1)){
    cuaProcedent=1;
    Quantum=QUANTUM_CUA1;
    return (tPCB *)Primer(pRDY1);
  } else {
    Quantum=QUANTUM_CUA2;
    cuaProcedent=2;
    return (tPCB *)Primer(pRDY2);
  }
}

// Realitza el canvi de context entre un nou proces i el que esta a Run
void dispatcher(tPCB *proc){
  // Visualitzem informacio per pantalla
  EscriureString(24,13,"Running",VERMELL);
  EscriureCar(24,21,procRun->IdProces,GROC);
  EscriureString(24,23,"CuaRDY",VERMELL);
  EscriureCar(24,30,48+cuaProcedent,GROC);
  EscriureString(24,32,"Prioritat",VERMELL);
  EscriureCar(24,42,48+procRun->Prioritat,GROC);
  
  proc->Estat=RUN;
  procRun=proc;
  Encuar(pRUN, proc);
  EOI();
  restaurarContext(proc->SS_SP);
}


// RSI que s'executa amb l'int 40h al fer una crida a sistema
// d'aquesta forma separem capa de nucli de la capa d'usuaris
void trap(void)
{
  static int ax;
  ax=obtenirAX();	// En el registre AX hem guardat l'identificador de la crida que es demanada
  procRun->SS_SP=salvarContext();
  // Al inipila (cridat desde CrearProces_K) al guardar flags activen el d'interrupcions.
  switch(ax)
  {
  case 1: CrearProces_K();
          break;
  case 2: DestruirProces_K();
          break;
  case 3: QuiSoc_K();
          break;
  case 4: DormirProces_K();
          break;
  case 5: IniSem_K();
          break;
  case 6: WaitS_K();
          break;
  case 7: SignalS_K();
          break;
  case 8: ElimSem_K();
          break;
  case 9: LlegirTeclat_K();
          break;
  case 10:EscriurePantalla_K();
          break;
  default: break;
 }
 EOI();
 restaurarContext(procRun->SS_SP);
}


// Programa principal
void main(void){
 tPCB *pcbAuxiliar;
 
 // Inicialitzacions
 iniCues();
 iniSemafors();
 BufferPantalla();
 // Creem proces Null, el qual crea el proces Shell tb
 CrearProcNull();

 flags=Inhibir();
 rsi_ck=CanviarVectInt(0x8,multiplexar);
 rsi_tr=CanviarVectInt(0x40,trap);
 rsi_kyb=CanviarVectInt(0x09,teclat);
 Desinhibir(flags);

 // Iniciem l'execuci agafant de la cua Ready 1
 pcbAuxiliar=(tPCB *)(Primer(pRDY1));
 pcbAuxiliar->Estat=RUN;
 Encuar(pRUN,(struct item *)pcbAuxiliar);
 procRun=pcbAuxiliar;
 restaurarContext(pcbAuxiliar->SS_SP);

}