/*
** Fitxer: nucli.c
** Data: Agost 2002
** Autors: Sergio Blanco Cuaresma
**         Jonathan Hernndez Velasco
**         Valentin Rey Gmez
** Descripci: Mdul amb les implementacins de les rutines (crides a sistema) a nivell nucli.
**
*/
#include "ec.h"
#include "estruct.h"
#include "pantalla.h"
#include "cues.h"
#include "main.h"
#include "salvrest.h"

///////////////////////
// CRIDES AL SISTEMA //
///////////////////////


// Crea un proces, si te una prioritat mes alta que el proces en execucio realitza apropiacio
int CrearProces_K (struct enviroment context,void (*codi)(), int prioritat) {
  static tPCB *pcbAuxiliar, *pcbNou;
  static WORD ss,sp;
  
  if (Buida(pFREE)){	// No hi han pcbs lliures
     context.ax=-1;	// Retornem error
     return (-1);
  }
  pcbNou = Primer(pFREE);
  pcbNou->IdProces = identificadorProc;
  identificadorProc=(identificadorProc+1)%256;  // Ens assegurem no sobrepassar el limit 255
  pcbNou->Estat = RDY;
  pcbNou->Prioritat = prioritat;
  pcbNou->QSencer[0] = TRUE;
  pcbNou->QSencer[1] = TRUE;
  ss = Segment(pcbNou->pila);
  sp = Desplacament(&(pcbNou->pila[MIDA_PILA-1]));
  pcbNou->SS_SP=iniPila(codi,ss,sp);
  // Mirem si hem de fer apropiacio
  if (procRun->Prioritat >= pcbNou->Prioritat) {
    Inserir (pRDY1, (struct item*)pcbNou, pcbNou->Prioritat);	// Quan es crear 1 proc sempre va a cua de ms prioritat
  } else {
    //procRun->SS_SP=salvarContext();  // No cal, ja s'ha guardat al trap
    pcbAuxiliar = (tPCB *)Primer(pRUN);
    pcbAuxiliar->Estat = RDY;
    pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
    pcbAuxiliar->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
    // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
    // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
    Inserir (pRDY1, (struct item*)pcbAuxiliar, pcbAuxiliar->Prioritat);
    cuaProcedent=1;
    Quantum = QUANTUM_CUA1;
    dispatcher(pcbNou);
  }
  context.ax=(identificadorProc-1);
  return (identificadorProc-1);
}


// ********************************************************************
// Treu de la cua que se li passa el PCB amb el identificador indicat
// Funciona igual que el extirpar pero no se li ha de passar tot el PCB
// Retorna l'element que s'ha tret
// ********************************************************************
tPCB *(Treure_de_cua(struct cua *c,int idProc))
{
 tPCB *post,*ant;

 if (c==NIL) return (NIL);
 post=(struct TipusPCB *)c->primer;
 if (post==NIL) return (NIL);

 ant=(struct TipusPCB *)c;
 while ((post!=(struct TipusPCB *)c) && ((post->IdProces)!=idProc))
 {
  ant=post;
  post=(struct TipusPCB *)post->Fil.seguent;
 }
 if (post==(struct TipusPCB *)c) return (NIL);

 ant->Fil.seguent=post->Fil.seguent;
 (post->Fil.seguent)->anterior=(struct item*)ant;
 post->Fil.seguent=NIL;
 post->Fil.anterior=NIL;
 return (post);
}


// Treu del sistema el proces indicat (buscant a totes les cues)
int DestruirProces_K(struct enviroment context,int Id)
{
 tPCB *pcbAuxiliar, *pcbExe;
 int i=0;

 // no es pot eliminar el proces null
 if (Id=='A') {
  context.ax=-2;
  return (-2);
 }

 if ((procRun->IdProces)==Id) // si el proces a destruir es el que esta a Run, el marquem com finalitzat
 {
  pcbAuxiliar=(tPCB *)Primer(pRUN);   // treiem el proces de la cua de Run
  Encuar(pFREE,(struct item *)pcbAuxiliar);       // i fiquem el pcb a la cua de pcb's lliures
  pcbAuxiliar->Estat=FREE;
  context.ax=0;
  dispatcher(scheduler());
 }
 else
 {
  // mirem si el proces esta ala cua de ready 
  pcbAuxiliar=Treure_de_cua(pRDY1,Id);
  if (pcbAuxiliar!=NIL)
  {
   Encuar(pFREE,(struct item *)pcbAuxiliar);
   context.ax=0;
   return (0);
  }
  // mirem si el proces esta ala cua de ready 
  pcbAuxiliar=Treure_de_cua(pRDY2,Id);
  if (pcbAuxiliar!=NIL)
  {
   Encuar(pFREE,(struct item *)pcbAuxiliar);
   context.ax=0;
   return (0);
  }
  // mirem si el proces esta a la cua de dormits
  pcbAuxiliar=Treure_de_cua(pSLEEP,Id);
  if (pcbAuxiliar!=NIL)
  {
   Encuar(pFREE,(struct item *)pcbAuxiliar);
   context.ax=0;
   return (0);
  }
  // mirem si el proces esta a la cua de bloquejats per teclat
  pcbAuxiliar=Treure_de_cua(pBLK,Id);
  if (pcbAuxiliar!=NIL)
  {
   Encuar(pFREE,(struct item *)pcbAuxiliar);
   context.ax=0;
   return (0);
  }
  // mirem si el proces esta a a cua del semafor de pantalla
  pcbAuxiliar=Treure_de_cua(SemaforPantalla.pBLK,Id);
  if (pcbAuxiliar!=NIL)
  {
   Encuar(pFREE,(struct item *)pcbAuxiliar);
   context.ax=0;
   return (0);
  }

  // mirem si el proces esta a la cua de algun semafor
  for (i=0;i<NSEM;i++){
   if (Semafors[i].id!=NSEM){
     pcbAuxiliar=Treure_de_cua(Semafors[i].pBLK,Id);
     if (pcbAuxiliar!=NIL)
     {
       Encuar(pFREE,(struct item *)pcbAuxiliar);
       context.ax=0;
       return (0);
     }
   }
  }
 }

 context.ax=-1;
 return (-1);
}


// Retorna l'identificador del proces que esta en execucio
int QuiSoc_K(struct enviroment context)
{
 int id;
 id=procRun->IdProces;
 context.ax=id;
 return (id);
}

// Treu de Run un proces i el fica a la cua d'adormits per n tics
void DormirProces_K(struct enviroment context,long NTics)
{
  static tPCB *pcbAuxiliar;

  // El treiem de Run per ficar-ho a cua sleep
  //procRun->SS_SP=salvarContext();  // No cal, ja s'ha guardat al trap
  pcbAuxiliar=(struct TipusPCB *)Primer(pRUN);     // treiem el proces de la cua de Run
  pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
  pcbAuxiliar->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
  pcbAuxiliar->Dormir=NTics;
  pcbAuxiliar->Estat=ASLEEP;
  Encuar(pSLEEP,(struct item *)pcbAuxiliar);
  // Agafem nou proces
  if (!Buida(pRDY1)){
    cuaProcedent=1;
    Quantum=QUANTUM_CUA1;
    dispatcher((tPCB *)Primer(pRDY1));
  } else {
    Quantum=QUANTUM_CUA2;
    cuaProcedent=2;
    dispatcher((tPCB *)Primer(pRDY2));
  }
}




// Inicialitza un semafor amb l'identificador i el valor indicat
int IniSem_K(struct enviroment context,int IdSemafor,int valor)
{
  WORD flags;
  int i, lliure=-1;

  flags=Inhibir();
  if (SemaforsLliures == 0) {
    Desinhibir(flags);		// No quedem semafors
    context.ax=-1;
    return(-1);
  }
  for (i=0;i<NSEM;i++){
    if (Semafors[i].id==IdSemafor) {  // L'identificador ja esta en s
      Desinhibir(flags);
      context.ax=-2;
      return(-2);
    }
    if ((Semafors[i].id==NSEM) && (lliure==-1)){
      lliure=i;
    }
  }
  Semafors[lliure].id=IdSemafor;
  Semafors[lliure].valor=valor;
  Inicialitzar_cua(Semafors[lliure].pBLK);
  SemaforsLliures--;
  Desinhibir(flags);
  context.ax=0;
  return(0);
}


// Bloqueja el proces que es en execucio si el semafor te valor 0
int WaitS_K(struct enviroment context,int IdSemafor)
{
  static tPCB *pcbAuxiliar;
  static WORD flags;
  static int i;

  flags=Inhibir();
  // busquem el semafor dins del vector de semafors
  for (i=0;i<NSEM;i++){
    if (Semafors[i].id==IdSemafor) {
      break;
    }
  }

  if (Semafors[i].id!=IdSemafor) {
   Desinhibir(flags);		// No existeix el semafor amb aquest identificador
   context.ax=-1;
   return(-1);
  }

  // si el semafor esta utilitzant-se
  if (Semafors[i].valor==0)
  {
   Desinhibir(flags);
   // El treiem de Run per ficar-ho a cua BLK
   //procRun->SS_SP=salvarContext();  // No cal, ja s'ha guardat al trap
   pcbAuxiliar=(struct TipusPCB *)Primer(pRUN);     // treiem el proces de la cua de Run
   pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
   pcbAuxiliar->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
   pcbAuxiliar->Estat=BLK;
   Encuar(Semafors[i].pBLK,(struct item *)pcbAuxiliar);
   // Agafem nou proces
   if (!Buida(pRDY1)){
     cuaProcedent=1;
     Quantum=QUANTUM_CUA1;
     dispatcher((tPCB *)Primer(pRDY1));
   } else {
     Quantum=QUANTUM_CUA2;
     cuaProcedent=2;
     dispatcher((tPCB *)Primer(pRDY2));
   }
  } else {
   Semafors[i].valor--;
   Desinhibir(flags);
  }
  context.ax=0;
  return(0);
}


// Augmenta el valor del semafor si no hi ha cap proces bloquejat a la seva cua.Sino, l'allibera
// comprovant si cal fer apropiacio
int SignalS_K(struct enviroment context,int IdSemafor)
{
  static WORD flags;
  static int i;
  static tPCB *pcbAuxiliar;
  static tPCB *pcbExe;

  flags=Inhibir();
  // busquem el semafor dins del vector de semafors
  for (i=0;i<NSEM;i++){
    if (Semafors[i].id==IdSemafor) {
      break;
    }
  }

  if (Semafors[i].id!=IdSemafor) {
   Desinhibir(flags);		// No existeix el semafor amb aquest identificador
   context.ax=-1;
   return(-1);
  }

 
  if (Buida(Semafors[i].pBLK)) {
    Semafors[i].valor++;
  } else {
    Desinhibir(flags);
    pcbAuxiliar=(tPCB *)Primer(Semafors[i].pBLK);
    if (procRun->Prioritat >= pcbAuxiliar->Prioritat) {		// No hem de fer apropiacio
        Inserir(pRDY1,(struct item *)pcbAuxiliar,pcbAuxiliar->Prioritat);
    } else {		// Hem de fer apropiacio
    	Desinhibir(flags);
        //procRun->SS_SP=salvarContext();  // No cal, ja s'ha guardat al trap
        pcbExe = (tPCB *)Primer(pRUN);
        pcbExe->Estat = RDY;
        pcbExe->QSencer[0] = pcbExe->QSencer[1];	// Historic d'execucions amb quantum complet
        pcbExe->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
        // Eleccio de la cua: Historic TRUE TRUE y FALSE TRUE queda descartat,
        // tan en el cas FALSE FALSE com TRUE FALSE ha d'anar a la cua 1 independenment de la procedencia
        Inserir (pRDY1, (struct item*)pcbExe, pcbExe->Prioritat);
        cuaProcedent=1;
        Quantum = QUANTUM_CUA1;
        dispatcher(pcbAuxiliar);
    }
  }
  Desinhibir(flags);
  context.ax=0;
  return(0);
}


// Elimina el semafor indicat deixant-lo com a lliure, si el semafor te processos bloquejats a la
// seva cua, retorna error i no l'elimina
int ElimSem_K(struct enviroment context,int IdSemafor)
{
  WORD flags;
  int i;

  flags=Inhibir();
  // busquem el semafor dins del vector de semafors
  for (i=0;i<NSEM;i++){
    if (Semafors[i].id==IdSemafor) {
      break;
    }
  }

  if (Semafors[i].id!=IdSemafor) {
   Desinhibir(flags);		// No existeix el semafor amb aquest identificador
   context.ax=-1;
   return(-1);
  }
  
  if (Buida(Semafors[i].pBLK)) {
    SemaforsLliures++;
    Semafors[i].id=NSEM;
    Desinhibir(flags);
    context.ax=0;
    return(0);
  } else {			// Encara hi han procesos pendents d'aquest semafor
    Desinhibir(flags);		
    context.ax=-2;
    return(-2); 
  }
}

// Llegeix del teclat tans caracters com els indicats i els guarda al buffer
int LlegirTeclat_K(struct enviroment context,char *Buffer,int longitud)
{
 static tPCB *pcbAuxiliar;
 static tPCB *procBLK;
 static int i;

 if (tecles>=longitud)
 {
   //Traspasem buffer al pcb
   for (i=0;i<longitud;i++) {
     Buffer[i]=bufferTeclat[i];
   }
   tecles=0;
 } else {
  // El treiem de Run per ficar-ho a cua BLK
  //procRun->SS_SP=salvarContext();  // No cal, ja s'ha guardat al trap
  pcbAuxiliar=(struct TipusPCB *)Primer(pRUN);     // treiem el proces de la cua de Run
  pcbAuxiliar->QSencer[0] = pcbAuxiliar->QSencer[1];	// Historic d'execucions amb quantum complet
  pcbAuxiliar->QSencer[1] = FALSE;			// False, no ha utilitzat tot el quantum
  pcbAuxiliar->Estat=BLK;
  pcbAuxiliar->llegir=longitud;
  pcbAuxiliar->buffer=Buffer;
  Encuar(pBLK,(struct item *)pcbAuxiliar);
  // Agafem nou proces
  if (!Buida(pRDY1)){
    cuaProcedent=1;
    Quantum=QUANTUM_CUA1;
    dispatcher((tPCB *)Primer(pRDY1));
  } else {
    Quantum=QUANTUM_CUA2;
    cuaProcedent=2;
    dispatcher((tPCB *)Primer(pRDY2));
  }
  // ATENCIO, VEURE AIXO:
  /*if (proc_act->idProces==1) Encuar(pcuaBlockKB,(struct item *)pcbExe);
  else Inserir(pcuaBlockKB,(struct item *)pcbExe,11);*/

 }
 context.ax=longitud;
 return (longitud);
}


// Escriu per pantalla el buffer, de tamany longitud a la fila i columna indicada amb l'atibut x
int EscriurePantalla_K(struct enviroment context,char *Buffer,int longitud,int fila,int columna,int atribut)
{
 int i;

 if (longitud>1) {
   WaitSPantalla();
 }
 i=0;
 while (i<longitud)
 {
  if (Buffer[i]!=13){		// No deixem imprimir els intros pq dibuixa una nota musical
    EscriureCar(fila,columna+i,Buffer[i],atribut);
  }
  i++;
 }
 if (longitud>1){
   SignalSPantalla();
 }
 context.ax=0;
 return 0;
}
